import copy
from contextlib import contextmanager

class ReadOnlyException(Exception):
    pass

class Definitions(object):
    def __init__(self, attributes):
        super(Definitions, self).__init__()
        self._attrs = {}
        self._fill_attributes(attributes)
        self._change_allowed = False
        self._backup_stack = []
        self._parent = None
    def _is_change_allowed(self):
        # pylint: disable=W0212
        return self._change_allowed or (self._parent is not None and self._parent._is_change_allowed())
    def _set_parent(self, parent):
        self._parent = parent
    def _fill_attributes(self, attrs):
        for key, value in attrs.items():
            if isinstance(value, dict):
                value = Definitions(value)
                # pylint: disable=W0212
                value._set_parent(self)
            self._attrs[key] = value
    def __getattr__(self, attr):
        if attr.startswith('_'):
            raise AttributeError(attr)
        try:
            return self._attrs[attr]
        except LookupError:
            raise AttributeError(attr)
    def __iter__(self):
        return iter(self._attrs)
    def __getitem__(self, item):
        return self._attrs[item]
    def __setitem__(self, item, value):
        if self._is_change_allowed() and item in self._attrs:
            self._attrs[item] = value
        else:
            raise ReadOnlyException("Cannot set {0}".format(item))
    def __setattr__(self, attr, value):
        if attr.startswith("_"):
            return super(Definitions, self).__setattr__(attr, value)
        self[attr] = value
    @contextmanager
    def __changing__(self):
        old_change_allowed = self._change_allowed
        self._backup_stack.append(copy.deepcopy(self._attrs))
        self._change_allowed = True
        try:
            yield
        finally:
            self._change_allowed = old_change_allowed
    def __restore__(self):
        self._attrs = self._backup_stack.pop(0)
    def __todict__(self):
        returned = {}
        for key in self:
            value = self[key]
            if isinstance(value, Definitions):
                value = value.__todict__()
            returned[key] = value
        return returned

def changing(defs):
    return defs.__changing__()
def restore(defs):
    return defs.__restore__()
def to_dict(defs):
    return defs.__todict__()
