import flux
from contextlib import contextmanager
import functools
import pyformance
from base64 import b64encode
import httplib
import requests
from requests.exceptions import RequestException
try:
    from requests.packages.urllib3.exceptions import HTTPError
except ImportError:
    # RPM version of requests remaps requests.packages.urllib3 to urllib3
    from urllib3.exceptions import HTTPError
import itertools
import logging
import socket
import uuid
import cjson
import json

from urlobject import URLObject as URL

from sentinels import NOTHING
from infinibox_sysdefs import latest as sysdefs

from ...config import config
from ..exceptions import (
    APICommandFailed,
    InvalidCommandOutput,
    APITransportFailure,
)
from .request_sender import RequestSender


class RAW_DATA(object):

    def __init__(self, data):
        super(RAW_DATA, self).__init__()
        self.data = data

    def __repr__(self):
        return "RAW({0!r})".format(self.data)

    def __str__(self):
        return "RAW({0})".format(self.data)

# Squash requests log by default
REQUEST_LOGGER = "requests"
logging.getLogger(REQUEST_LOGGER).setLevel(logging.WARNING)

API_LOGGER = "infinipy.api"
DEBUG = logging.getLogger(API_LOGGER).debug
RESPONSE_DEBUG = logging.getLogger("infinipy.api.response").debug
_OK_STATUSES = set([httplib.OK, httplib.CREATED])

pyformance.registry._global_registry = pyformance.registry.RegexRegistry(
    config.root.defaults.pyformance_pattern)


def measure(path):
    return pyformance.timer(path).time(path=path)


class HasRelativeAPIMixin(object):

    def get_sub_api(self, path):
        return RelativeAPI(self, path)


class AutoRetryMixin(object):
    def __init__(self):
        self._autoretry_exceptions = {}

    def add_autoretry_exception(self, should_retry_callback, max_retries=1):
        DEBUG("Add auto-retry context {0} for {1} retries".format(should_retry_callback, max_retries))
        self._autoretry_exceptions[should_retry_callback] = max_retries

    def remove_autoretry_exception(self, should_retry_callback):
        DEBUG("Remove auto-retry context {0}".format(should_retry_callback))
        del self._autoretry_exceptions[should_retry_callback]

    @staticmethod
    def handle_autoretry(func):
        @functools.wraps(func)
        def wrapper(self, *args, **kwargs):
            retry_count = dict(self._autoretry_exceptions)
            while True:
                try:
                    return func(self, *args, **kwargs)
                except Exception as e:
                    retry = False
                    for should_retry_callback, max_retries in self._autoretry_exceptions.iteritems():
                        if retry_count.get(should_retry_callback, 0) == 0:
                            raise
                        if should_retry_callback(e):
                            retry = True
                            DEBUG("Auto retry API {0} of {1}: {2} was caught by {3}".format(max_retries - retry_count[should_retry_callback] + 1, max_retries, e, should_retry_callback))
                            retry_count[should_retry_callback] -= 1
                    if retry:
                        flux.current_timeline.sleep(config.root.defaults.system_api_auto_retry_sleep)
                    else:
                        raise
        return wrapper


class RelativeAPI(RequestSender, HasRelativeAPIMixin):

    def __init__(self, parent, path):
        super(RelativeAPI, self).__init__()
        self.api_url = parent.api_url
        self._parent = parent
        self._url = path

    def _request(self, method, path, *args, **kwargs):
        return self._parent._request(method, _join(self._url, path), *args, **kwargs)

def _build_delegating_method(name):
    def _delegate(self, *args, **kwargs):
        return getattr(self._parent, name)(*args, **kwargs)
    _delegate.__name__ = name
    return _delegate

for _delegated_method_name in [
        "get_command_id_context",
        "get_current_user",
        "get_login_context",
        "get_request_timeout",
        "get_unapproved_context",
        "get_username_password",
        "get_autoretry_context",
        "add_autoretry_exception",
        "remove_autoretry_exception",
        "set_request_timeout",
        "set_source_identifier",
        "set_username_password",
]:
    setattr(RelativeAPI, _delegated_method_name,
            _build_delegating_method(_delegated_method_name))


class RestAPIBase(RequestSender, HasRelativeAPIMixin, AutoRetryMixin):

    def __init__(self, system, username=None, password=None, ssl_cert=None):
        RequestSender.__init__(self)
        HasRelativeAPIMixin.__init__(self)
        AutoRetryMixin.__init__(self)
        self.system = system
        command_id_formatter = '{0}{{0}}'.format(uuid.uuid1()).format
        self._command_id_generator = itertools.imap(
            command_id_formatter, itertools.count())
        self._custom_command_id = None
        self._headers = {
            "Accept": "application/json",
        }
        self._timeout = config.root.defaults.system_api_timeout_seconds
        self.set_source_identifier(config.root.defaults.system_api_identifier)
        self.set_username_password(
            username,
            password
        )
        self._username_password_set = False
        self._ssl_cert=ssl_cert

    def set_request_timeout(self, timeout_seconds):
        self._timeout = timeout_seconds

    def get_request_timeout(self):
        return self._timeout

    def set_source_identifier(self, identifier):
        self._headers[
            sysdefs.api.headers.infinidat_source_header_name] = identifier

    def get_username_password(self):
        return self._username, self._password

    _current_user_object = None
    def get_current_user(self):
        if self._current_user_object is None:
            self._current_user_object = self.system.login()
        return self._current_user_object

    def set_username_password(self, username, password):
        if username is not None and password is None:
            raise RuntimeError("No password specified")
        self._username, self._password = username, password
        self._current_user_object = None
        if self._username is None:
            self._headers.pop("Authorization", None)
        else:
            self._headers["Authorization"] = self._get_authorization_string()
        self._username_password_set = True

    @contextmanager
    def get_login_context(self, *args, **kwargs):
        old_set_flag = self._username_password_set
        try:
            with super(RestAPIBase, self).get_login_context(*args, **kwargs):
                yield
        finally:
            self._username_password_set = old_set_flag

    def _get_authorization_string(self):
        return 'Basic ' + b64encode(('%s:%s' % (self._username, self._password)).encode('latin1')).strip().decode('latin1')

    @AutoRetryMixin.handle_autoretry
    def _request(self, method, path, data, headers, raw_json_response):
        path = URL(path)
        if self._approved and method != "GET":
            path = path.add_query_param("approved", "true")

        address_string = self.system.api.api_url.netloc

        DEBUG("%s <-- %s %s", address_string, method, path)
        header_dict = self._get_header_dict(method)
        header_dict.update(headers)
        ctx = RequestContext(self, method, path, data,
                             header_dict, raw_json_response)
        if ctx.sent_data is not None:
            assert isinstance(ctx.sent_data, str)
            DEBUG("%s <-- DATA: %r", address_string, _hide_password(ctx.sent_data))
        timer = measure(path)
        with self.system.callbacks.api_command_context.get_context(command=ctx):
            self._send_request_get_response(ctx)
            ctx.elapsed = timer.stop()
        RESPONSE_DEBUG("%s %s%s --> %s %.04fs %s/%s %r", method, address_string, path, ctx.status, ctx.elapsed,
              ctx.num_objects_returned, ctx.num_objects_total, ctx.output)
        self._check_response(ctx)
        return ctx

    def _send_request_get_response(self, ctx):
        raise NotImplementedError()  # pragma: no cover

    def _check_response(self, ctx):
        if ctx.status not in _OK_STATUSES:
            raise APICommandFailed(ctx)

    def _get_header_dict(self, method):
        returned = dict(self._headers)
        if method != 'GET':
            returned[
                sysdefs.api.headers.infinidat_command_id_header_name] = self._get_command_id()
        return returned

    def _get_command_id(self):
        if self._custom_command_id is not None:
            return self._custom_command_id
        return next(self._command_id_generator)

    @contextmanager
    def get_command_id_context(self, command_id):
        old_custom_command_id = self._custom_command_id
        self._custom_command_id = command_id
        try:
            yield
        finally:
            self._custom_command_id = old_custom_command_id

    _approved = True

    @contextmanager
    def get_unapproved_context(self):
        prev_approved = self._approved
        self._approved = False
        try:
            yield
        finally:
            self._approved = prev_approved

    def _check_response(self, ctx):
        if ctx.status not in _OK_STATUSES:
            raise APICommandFailed(ctx)

    @contextmanager
    def get_autoretry_context(self, should_retry_callback, max_retries=1):
        self.add_autoretry_exception(should_retry_callback, max_retries)
        try:
            yield
        finally:
            self.remove_autoretry_exception(should_retry_callback)


def _join(*parts):
    returned = ""
    for part in parts:
        if not part:
            continue
        if returned.endswith("/") and part.startswith("/"):
            part = part[1:]
        elif returned and (not returned.endswith("/") or not part.startswith("/")):
            returned += "/"
        returned += part
    return returned


class RequestContext(object):

    metadata = num_objects_total = num_objects_returned = None

    def __init__(self, sender, method, path, sent_data, headers, raw_json_response=False):
        super(RequestContext, self).__init__()
        self.sender = sender
        self.method = method
        self.path = path
        self.sent_object = sent_data
        self.sent_headers = headers
        self.raw_json_response = raw_json_response
        self.full_url = self.sender.system.api.api_url + self.path
        if self.sent_object is NOTHING:
            self.sent_data = None
        else:
            if isinstance(self.sent_object, RAW_DATA):
                self.sent_data = self.sent_object.data
            else:
                self.sent_headers["Content-Type"] = "application/json"
                self.sent_data = json.dumps(
                    self.sent_object, ensure_ascii=False)

        if isinstance(self.sent_data, unicode):
            self.sent_data = self.sent_data.encode("utf-8")

        self.output = NOTHING
        self.raw_output = self.received_headers = None
        self.status = self.reason = self.status_and_reason = None

    def __repr__(self):
        return "<{0} {1} --> {2}>".format(self.method, self.path, self.status or "?")

    def set_response(self, raw_output, headers, status, reason):
        self.raw_output = raw_output
        self.received_headers = headers
        self.status = status
        self.reason = reason
        self.status_and_reason = "{0} {1}".format(self.status, self.reason)
        self.success = self.status in _OK_STATUSES
        if self._is_json_response():
            try:
                self.output = cjson.decode(self.raw_output.decode("utf-8"))
            except cjson.DecodeError:
                DEBUG("Cannot parse output from API call", exc_info=True)
                raise InvalidCommandOutput(self)
        elif self.success:
            if self._has_expected_json_response():
                DEBUG("We got a success response, but content is not JSON")
                DEBUG("Response headers are: %r", self.received_headers)
                raise InvalidCommandOutput(self)
            else:
                self.output = self.raw_output.decode("utf-8")

        if self.success:
            if isinstance(self.output, dict):
                self.metadata = self.output.get("metadata")
                if self.metadata is None:
                    raise InvalidCommandOutput(self)
                try:
                    self.num_objects_returned = len(self.output.get("result"))
                except (ValueError, TypeError):
                    pass
                self.num_objects_total = self.metadata.get("number_of_objects")

                if not self.raw_json_response and isinstance(self.output, dict):
                    self.output = self.output["result"]

    def _is_json_response(self):
        return "application/json" in self.received_headers.get("content-type", "")

    def _has_expected_json_response(self):
        return "application/json" in self.sent_headers.get("Accept", "")

class RestAPI(RestAPIBase):

    def __init__(self, system, api_url, resolve_hostname=True, **kwargs):
        super(RestAPI, self).__init__(system, **kwargs)
        self.api_url = api_url

    def get_user(self):
        raise NotImplementedError()  # pragma: no cover

    def _send_request_get_response(self, ctx):
        try:
            response = requests.request(ctx.method, ctx.full_url,
                                        headers=ctx.sent_headers, data=ctx.sent_data,
                                        timeout=self.get_request_timeout(), cert=self._ssl_cert,
                                        verify=self._ssl_cert is not None)
        except (socket.error, HTTPError, RequestException) as e:
            DEBUG("Socket error during request sending", exc_info=True)
            raise APITransportFailure(ctx, str(e), exception=e)
        ctx.set_response(response.content, response.headers,
                         response.status_code, response.reason)


class _hide_password(object):

    def __init__(self, s):
        super(_hide_password, self).__init__()
        self.s = s

    def __repr__(self):
        returned = self.s
        try:
            original_json = json.loads(returned)
            if 'password' in original_json:
                original_json['password'] = '*' * len(original_json['password'])
                returned = json.dumps(original_json)
        except Exception:
            pass
        return returned
