from contextlib import contextmanager
from capacity import Capacity, byte
from infi.pyutils.lazy import cached_method
from infinibox_sysdefs import get_config_parameter_type


class Configuration(object):

    def __init__(self, system):
        super(Configuration, self).__init__()
        self._system = system
        self._deployment_id = self._system.api.get('system')['deployment_id']
        self._needs_reverting = {}

    def get_deployment_id(self):
        return self._deployment_id

    def get_limit(self, name):
        return self.get_limits_dict()[name]
    _cached_limits_dict = None

    def get_limits_dict(self):
        if self._cached_limits_dict is None:
            self._cached_limits_dict = dict((key, int(value))
                                            for key, value in self._system.api.get("config/limits").iteritems())
        return self._cached_limits_dict

    def get_parameter(self, path):
        returned = self._deserialize_from_system(
            path, self._system.api.get("config/" + path))
        return returned

    def set_parameter(self, path, value):
        if path.startswith("limits"):
            # simple cache invalidation for limits
            self._cached_limits_dict = None
        if path not in self._needs_reverting:
            self._needs_reverting[path] = self.get_parameter(path)
        return self._system.api.put("config/" + path,
                                    data=self._serialize_to_system(path, value))

    def revert_parametrs(self):
        for path in list(self._needs_reverting):
            self.revert_parameter(path)

    def revert_parameter(self, path):
        if path in self._needs_reverting:
            self.set_parameter(path, self._needs_reverting[path])
            self._needs_reverting.pop(path)

    @contextmanager
    def get_changed_parameter_context(self, path, value):
        old_value = self.get_parameter(path)
        self.set_parameter(path, value)
        try:
            yield
        finally:
            self.set_parameter(path, old_value)

    def _deserialize_from_system(self, path, value):
        assert not path.endswith("/") and "//" not in path
        path_type = get_config_parameter_type(path)
        if path_type is Capacity:
            return int(value) * byte
        return path_type(value)

    def _serialize_to_system(self, path, value):
        assert not path.endswith("/") and "//" not in path
        path_type = get_config_parameter_type(path)
        if path_type is Capacity:
            value = value // byte
        return value
