from capacity import byte
from .taggable_object import TaggableObject
from .physical_object import PhaseableObject
from .exceptions import APICommandFailed
from infinibox_sysdefs import latest as sysdefs

DRIVE_STATES = sysdefs.enums.enclosures.drives.states


class Drive(PhaseableObject, TaggableObject):
    states = DRIVE_STATES
    _base_api_path = "components/racks/{0}/enclosures/{1}/drives/{2}"

    def __init__(self, system, enclosure, drive_info):
        super(Drive, self).__init__(system)
        self._enclosure = enclosure
        self._enclosure_index = drive_info['enclosure_index']
        assert enclosure.get_id() == self._enclosure_index
        self._drive_index = drive_info['drive_index']
        # not zero based...
        assert self._enclosure_index != 0
        assert self._drive_index != 0

    def _get_tag_id(self):
        return "{0}:{1}".format(self._enclosure_index, self._drive_index)

    def _get_key(self):
        return (self._system, type(self), self._enclosure_index, self._drive_index)

    def _get_api_path(self, cmd="", *args, **kwargs):
        return super(Drive, self)._get_api_path(cmd, self.get_system().get_rack_id(), self.get_enclosure_index(), self.get_drive_index(), *args, **kwargs)

    def is_running(self):
        return self.is_active()

    def get_paths(self):
        return self._get_underlying_object()['nodes_access']

    def is_node_path_active(self, node):
        nodes_access = self.get_paths()
        return nodes_access[node.get_id() - 1] if len(nodes_access) >= node.get_id() else False

    def get_serial_number(self):
        obj = self._get_underlying_object()
        return obj.get('serial_number', None)

    def get_product_id(self):
        obj = self._get_underlying_object()
        return obj.get('product_id', None)

    def get_capacity(self):
        return self._get_underlying_object()['bytes_capacity'] * byte

    def _get_drives_on_node(self, node):
        drives = []
        for enclosure in self._system.api.get('components/racks/1/enclosures'):
            for drive in enclosure['drives']:
                nodes_access = drive['nodes_access']
                if len(nodes_access) >= node.get_id() and nodes_access[node.get_id() - 1]:
                    drives.append(drive)
        return drives

    def _get_drive_on_node(self, node):
        if self.is_node_path_active(node):
            return self
        return None

    def _safe_to_kill_query(self):
        try:
            internal_api_path = "core/core_api/SAFE_TO_KILL_DRIVE_QUERY/{0}/{1}".format(
                self._enclosure_index, self._drive_index)
            return self.get_system().internal_api.get(internal_api_path)
        # internal API isn't well protected. Assume a fail means no.
        except APICommandFailed:
            return {'possible_to_phaseout': False,
                    'safe': False}

    def is_possible_to_phaseout(self):
        return self._safe_to_kill_query()['possible_to_phaseout']

    def is_redundant(self):
        return self._safe_to_kill_query()['safe']

    def _get_state_api_path(self):
        return "{0}/state".format(self._get_drive_api_path())

    def __repr__(self):
        return "<{0} {1!r}:E{2}D{3}>".format(type(self).__name__, self.get_system(), self._enclosure_index, self._drive_index)

    def get_enclosure(self):
        return self._enclosure

    def get_enclosure_index(self):
        return self._enclosure_index

    def get_drive_index(self):
        return self._drive_index

    def test(self):
        self._system.api.post(self._get_api_path("/test"))
