from munch import Munch
from itertools import count
from urlobject.query_string import QueryString

default_event_page_size = 1000
default_max_events = 10 ** 6
max_events_page_size = 1000

class Events(object):

    def __init__(self, system):
        super(Events, self).__init__()
        self._system = system

    def _get_events(self, query="", page_size=default_event_page_size, max_events=None):
        query = "events?{query}&page_size={page_size}&".format(query=query, page_size=page_size) + "page={page}"
        returned = []
        total_num_objects = max_events
        for page_index in count(1):
            result = self._system.api.get(query.format(page=page_index), raw_json_response=True)
            if total_num_objects is None:
                total_num_objects = result["metadata"]["number_of_objects"]
            events = [Munch(e) for e in result["result"]]
            returned.extend(events)
            if len(events) < page_size or len(returned) >= total_num_objects:
                break
        del returned[total_num_objects:]
        return returned

    def create_custom_event(self, level="INFO", visibility="INFINIDAT", description_template="custom event", data=None):
        if not data:
            data = []
        return self._system.api.post("events/custom", data=dict(
            level=level,
            visibility=visibility,
            description_template=description_template,
            data=data,
        ))

    def get_events(self, *args, **kwargs):
        return self._get_events(self.get_event_filter_string(*args, **kwargs))

    def get_event(self, event_id):
        return Munch(self._system.api.get("events/{0}".format(event_id)))

    def _normalize_sort_field(self, field_name, direction):
        direction = direction.lower()
        assert direction in ("-", "+", "asc", "desc")
        if direction in ("-", "desc"):
            field_name = "%{0:02X}{1}".format("-", field_name)
        return field_name

    def get_event_filter_string(self, min_event_id=None, code=None, sort="+"):
        filter_expressions = []

        query = "sort={0}".format(self._normalize_sort_field("id", sort))
        if min_event_id is not None:
            filter_expressions.append("id=ge:{0}".format(min_event_id))
        if code is not None:
            filter_expressions.append("code=eq:{0}".format(code))
        if filter_expressions:
            query += "&{0}".format("&".join(filter_expressions))
        return query

    def get_last_events(self, num, reversed=False):
        returned = self._get_events(query="sort=%2did", page_size=min(max_events_page_size, num), max_events=num)
        del returned[num:]
        if not reversed:
            returned.reverse()
        return returned
    def get_last_event(self):
        last_events = self.get_last_events(1)
        if last_events:
            return last_events[0]

    def custom_event_query(self, **kwargs):
        return self._get_events(**kwargs)
    def _get_events_types(self):
        return self._system.api.get("events/types")
    def get_codes(self):
        return self._get_events_types()['codes']
    def get_levels(self):
        sorted_levels = sorted(self._get_events_types()['levels'], key=lambda d: d['value'])
        return [level_info['name'] for level_info in sorted_levels]
    def get_visibilities(self):
        return self._get_events_types()['visibilities']
    def get_reporters(self):
        return self._get_events_types()['reporters']

    def _get_anti_flooding_path(self):
        return "mgmt/events.flooding_detector.enabled"
    def _set_anti_flooding_value(self, value):
        self._system.configuration.set_parameter(self._get_anti_flooding_path(), value)
    def is_anti_flooding_enabled(self):
        return self._system.configuration.get_parameter(self._get_anti_flooding_path())
    def disable_anti_flooding(self):
        return self._set_anti_flooding_value(False)
    def enable_anti_flooding(self):
        return self._set_anti_flooding_value(True)
