from .physical_object import PhysicalObject, StatefulObject

from infinibox_sysdefs import latest as sysdefs

NODE_DRIVE_STATES = sysdefs.enums.nodes.drives.states
NODE_DRIVE_TYPES = sysdefs.enums.nodes.drives.types
NODE_DRIVE_BAY_STATES = sysdefs.enums.nodes.drive_bays.states


class LocalDrive(PhysicalObject):
    states = NODE_DRIVE_STATES
    types = NODE_DRIVE_TYPES

    def __init__(self, node, drive_config):
        super(LocalDrive, self).__init__(node.get_system())
        self._node = node
        self._id = drive_config['id']
        self._index = drive_config['drive_index']
        self._serial = drive_config['serial_number']
        self._model = drive_config['model']
        self._vendor = drive_config['vendor']
        self._firmware = drive_config['firmware']
        self._type = drive_config['type']

    def _get_key(self):
        return (self._node, type(self), (self._index))

    def get_index(self):
        return self._index

    def get_node(self):
        return self._node

    def get_serial_number(self):
        return self._serial

    def get_model(self):
        return self._model

    def get_vendor(self):
        return self._vendor

    def get_firmware(self):
        return self._firmware

    def is_ssd_caching_active(self):
        if self.is_ssd():
            ssd_drive_query = self.get_system().internal_api.post("core/core_api/SSD_DRIVE_QUERY", data={"node_id":self.get_node().get_id()})
            for ssd_drive in ssd_drive_query["ssd_drive_list"]:
                if ssd_drive["drive_index"] == self.get_index():
                    if ssd_drive["state"] == "ACTIVE":
                        return True
                    else:
                        return False
        return False

    def is_ssd(self):
        return self.get_type() == self.types.ssd

    def get_type(self):
        return self.types.get(self._type)

    def _get_api_path(self):
        return "{0}/drives/{1}".format(self._node._get_api_path(), self._id)

    def __repr__(self):
        return "<{0}:{1} <{2}> on {3!r}>".format(type(self).__name__, self._index, self._serial, self._node)

class LocalDriveBay(StatefulObject):
    states = NODE_DRIVE_BAY_STATES

    def __init__(self, node, index):
        super(LocalDriveBay, self).__init__()
        self._node = node
        self._index = index
        self._drive = None
        self.get_drive()  # Attempt to populate upon init

    @property
    def index(self):
        return self.get_index()

    @property
    def drive(self):
        return self._drive

    def get_index(self):
        return self._index

    def get_drive(self):
        if self.drive is None:
            for drive in self._node.get_local_drives():
                if drive.get_index() == self.index:
                    self._drive = drive
                    break
        return self.drive

    def get_state(self):
        if self.drive is not None:
            return self.states.populated
        return self.states.empty

    def __repr__(self):
        return "<{0}:{1} on {2!r}>".format(type(self).__name__, self._index, self._node)
