from .fiber_port import FiberPort
from .local_drive import LocalDrive, LocalDriveBay
from .physical_object import PhysicalObject, PhaseableObject
import logging

from infinibox_sysdefs import latest as sysdefs

NODE_STATES = sysdefs.enums.nodes.states
SERVICE_STATES = sysdefs.enums.nodes.services.states
SERVICE_ROLES = sysdefs.enums.nodes.services.roles

_logger = logging.getLogger()

class Node(PhaseableObject):
    states = NODE_STATES
    _base_api_path = "components/racks/{0}/nodes/{1}"
    def __init__(self, system, node_config):
        super(Node, self).__init__(system)
        self._index = node_config['id']
        assert self._index != 0, "Node id's must be 1-based!"
        self._fiber_ports = [FiberPort(self, fc_config) for fc_config in node_config['fc_ports']]
        self._local_drives = [LocalDrive(self, drive_config) for drive_config in node_config['drives']]
        self._local_drive_bays = [LocalDriveBay(self, location) for location in range(1, 17)]  # TODO - get drive bays according to server type
        self._is_virtual = any(drive.get_vendor().lower() == "vmware" for drive in self._local_drives \
                               if drive.get_vendor() is not None)

    def is_virtual(self):
        return self._is_virtual

    def _get_key(self):
        return (self._system, self._index)

    def __repr__(self):
        return "<Node {0} of {1}>".format(self._index, self._system)
    def get_id(self):
        return self._index
    def get_fiber_ports(self):
        """
        :rtype: a list of :class:`infinipy.system.fiber_port.FiberPort` objects
        """
        return self._fiber_ports
    def get_local_drives(self):
        """
        :rtype: a list of :class:`infinipy.system.local_drive.LocalDrive` objects
        """
        return self._local_drives
    def get_local_drive_bays(self):
        """
        :rtype: a list of :class:`infinipy.system.local_drive.LocalDriveBay` objects
        """
        return self._local_drive_bays
    def get_active_ssd_drives(self):
        return [d for d in self.get_local_drives() if d.is_ssd_caching_active()]
    def _get_api_path(self, cmd="", *args, **kwargs):
        return super(Node, self)._get_api_path(cmd, self.get_system().get_rack_id(), self.get_id(), *args, **kwargs)
    def _update_local_drive(self, index):
        drives_config = self._get_underlying_object()['drives']
        updated_drive = [LocalDrive(self, drive_config) for drive_config in drives_config if int(drive_config['drive_index'])==int(index)]
        assert len(updated_drive) == 1
        for e, old_drive in enumerate(self._local_drives):
            if int(old_drive.get_index()) == int(index):
                self._local_drives[e] = updated_drive[0]
                break
        else:
            self._local_drives.append(updated_drive[0])
    def is_running(self):
        return self.is_active()
    def get_service(self, name):
        return Service(self, name)
    def get_core_service(self):
        return self.get_service('core')
    def get_management_service(self):
        return self.get_service('mgmt')
    def get_platform_service(self):
        return self.get_service('platform')

class Service(PhysicalObject):
    states = SERVICE_STATES
    roles = SERVICE_ROLES
    _base_api_path = "components/racks/{0}/nodes/{1}/services/{2}"
    def __init__(self, node, name):
        super(Service, self).__init__(node.get_system())
        self._node = node
        self._name = name
    def _get_api_path(self, cmd="", *args, **kwargs):
        return super(Service, self)._get_api_path(cmd, self.get_system().get_rack_id(), self._node.get_id(), self._name, *args, **kwargs)
    def _get_key(self):
        return (self._node, self._name)
    def __repr__(self):
        return "<{0} Service of {1}>".format(self._name, self._node)
    def get_role(self):
        return self._get_underlying_object()['role']
    def is_master(self):
        return self.get_role() == self.roles.master
    def is_secondary(self):
        return self.get_role() == self.roles.secondary
    def is_member(self):
        return self.get_role() == self.roles.member
    def get_number_of_respawns(self):
        return self._get_underlying_object()['number_of_respawns']
    def get_name(self):
        return self._name
    def get_node(self):
        return self._node
    def get_service_cluster(self):
        return self._system.get_service_cluster(self._name)
