from .system_object import NewAPICollectionSystemObject
import sys
import copy

try:
    from collections import OrderedDict
except ImportError:
    from ordereddict import OrderedDict

_FILTER_KEYS = ['event_code', 'event_level', 'event_visibility', 'include_events', 'exclude_events']
class NotificationRule(NewAPICollectionSystemObject):
    @classmethod
    def create(cls, system, name, target, **kwargs):
        data = dict(
                name=name,
                target_id=target.get_id(),
                **kwargs
            )
        return cls._create_from_post(system, data)
    @classmethod
    def get_collection_path(cls):
        return "notifications/rules"
    def set_target(self, target):
        return self.update_field('target_id', target.get_id())
    def get_target(self):
        return self._system.objects.NotificationTarget.get(id=self.get_field('target_id'))
    def get_target_parameters(self):
        return self.get_field('target_parameters')
    def set_target_parameters(self, data):
        return self.update_field("target_parameters", data)
    def set_filter(self, notification_filter):
        self.update_fields(**notification_filter.to_dict())
    def get_filter(self):
        return NotificationFilter(**self.get_fields(*self.get_filter_keys()))
    def get_filter_keys(self):
        return _FILTER_KEYS[:]
    def get_recipients(self):
        return self.get_target_parameters()["recipients"]
    def set_recipients(self, recipients):
        return self.set_target_parameters({"recipients":recipients})
    def add_recipient(self, recipient):
        return self.set_recipients(self.get_recipients() + [recipient])
    def set_visibility(self, new_visibility):
        return self.update_field("event_visibility", new_visibility)
    def get_visibility(self):
        return self.get_field("event_visibility")

    # old methods
    def delete_recipient(self, recipient):
        raise NotImplementedError("Deprecated - use set_recipients")
    def add_filter(self, email_filter):
        raise NotImplementedError("Deprecated")
    def get_filters(self):
        raise NotImplementedError("Deprecated")
    def set_filters(self, email_filters):
        raise NotImplementedError("Deprecated")
    def delete_filter(self, email_filter):
        raise NotImplementedError("Deprecated")


class NotificationFilter(object):
    def __init__(self, **kwargs):
        super(NotificationFilter, self).__init__()
        self._data = OrderedDict()
        for key in _FILTER_KEYS:
            self._data[key] = kwargs[key]
    def to_dict(self):
        return copy.deepcopy(self._data)

    def filter_events(self, events):
        """ filters a collection of events according to filter values and returns
            a collection of events matching the filter 
        """
        returned_events = []
        for event in events:
            event_code = event.get_code()
            if event_code == self._data['event_code']:
                returned_events.append(event)
                continue
            if event_code in self._data['exclude_events']:
                continue
            if event_code in self._data['include_events'] or all([event.get_attribute('visibility') in self._data['event_visibility'],
                                                                  event.get_attribute('level') in self._data['event_level']]):
                returned_events.append(event)
        return returned_events
    def __eq__(self, other):
        if not isinstance(other, NotificationFilter):
            return False
        return self.to_dict() == other.to_dict()

    def __hash__(self):
        return hash(tuple(["_".join(v) if isinstance(v, list) else v for v in self._data.values()]))
