import types
import functools
from infi.pyutils import HashableByKey
from .mixins import ExtensibleMixin
from .exceptions import APICommandFailed

def repartial(func, *parameters, **kparms):
    @functools.wraps(func)
    def wrapped(self, *args, **kw):
        kw.update(kparms)
        return func(self, *(args + parameters), **kw)
    return wrapped

class Meta(type):
    def __new__(cls, clsname, bases, attrs):
        if 'states' in attrs:
            for s in attrs['states']:
                func_name = 'is_{0}'.format(str(s).lower().replace(' ', '_'))
                if not func_name in attrs:
                    attrs[func_name] = repartial(StatefulObject._is_state, state=s)
        return super(Meta, cls).__new__(cls, clsname, bases, attrs)
    
class StatefulObject(ExtensibleMixin):
    __metaclass__ = Meta
    def __init__(self):
        super(StatefulObject, self).__init__()
    def _is_state(self, state):
        return self.get_state() == state
    def get_state(self):
        return self._get_underlying_object()['state']

class PhysicalObject(HashableByKey, StatefulObject):
    def __init__(self, system):
        super(PhysicalObject, self).__init__()
        self._system = system
    def get_system(self):
        return self._system
    def _get_underlying_object(self):
        return self._system.api.get(self._get_api_path())
    def _get_api_path(self, cmd="", *args, **kwargs):
        return (self._base_api_path + cmd).format(*args, **kwargs)

def phase_retry(e):
    return isinstance(e, APICommandFailed) and e.ctx.status == 503 and e.code in ['STATUS_RETRY', 'STATUS_IGNORED']

class PhaseableObject(PhysicalObject):
    def phase_out(self):
        with self._system.api.get_autoretry_context(phase_retry, max_retries=3):
            self._system.api.post(self._get_api_path("/phase_out"))
    def phase_in(self):
        with self._system.api.get_autoretry_context(phase_retry, max_retries=3):
            self._system.api.post(self._get_api_path("/phase_in"))
