import httplib
from capacity import byte
from munch import Munch

from ..config import config
from .exceptions import APICommandFailed
from .defaults import GENERATE_NAME, has_defaults
from .system_object import CollectionSystemObject
from .volume import Volume
from .host import Host
from .cluster import Cluster
from .utils import (
    convert_size_for_system,
    convert_size_from_system,
    )
from infinibox_sysdefs import latest as sysdefs

POOL_STATES = sysdefs.enums.pools.states

class Pool(CollectionSystemObject):
    CAPACITY_ATTRIBUTES = ['virtual_capacity', 'physical_capacity', 'reserved_capacity', 'max_extend',
                           'free_physical_space', 'free_virtual_space', 'allocated_physical_space']
    states = POOL_STATES
    @classmethod
    def _get_query_all_command(cls):
        return 'pool.query'
    @classmethod
    @has_defaults
    def create(cls, system, name=GENERATE_NAME, virtual_capacity=None, physical_capacity=None, **kwargs):
        if "soft_capacity" in kwargs or "hard_capacity" in kwargs:
            raise NotImplementedError("Hard/soft capacity are no longer supported by Infinibox. Use physical/virtual capacity definitions instead")  # pragma: no cover
        if physical_capacity is None:
            physical_capacity = config.root.defaults.pool.size
        if virtual_capacity is None:
            virtual_capacity = config.root.defaults.pool.size
        max_extend = kwargs.get('max_extend', None)
        if max_extend is not None:
            kwargs['max_extend'] = convert_size_for_system(max_extend)
        ctx = Munch(result=None)
        with system.callbacks.pool_creation_context.get_context(name=name, virtual_capacity=virtual_capacity, physical_capacity=physical_capacity, result=ctx):
            result = system.api.post('pools', dict(name=name,
                                                   virtual_capacity=convert_size_for_system(virtual_capacity),
                                                   physical_capacity=convert_size_for_system(physical_capacity),
                                                   **kwargs
                                                   ))
            ctx.result = cls(system, result['id'])
        return ctx.result
    # # attributes
    def get_virtual_capacity(self):
        return self._get_size_field('virtual_capacity')
    def set_virtual_capacity(self, virtual_capacity):
        ctx = Munch(result=None)
        with self.get_system().callbacks.pool_resize_context.get_context(pool=self, virtual_capacity=virtual_capacity, physical_capacity=None, result=ctx):
            return self._set_size_field('virtual_capacity', virtual_capacity)
    def get_physical_capacity(self):
        return self._get_size_field('physical_capacity')
    def set_physical_capacity(self, physical_capacity):
        ctx = Munch(result=None)
        with self.get_system().callbacks.pool_resize_context.get_context(pool=self, virtual_capacity=None, physical_capacity=physical_capacity, result=ctx):
            return self._set_size_field('physical_capacity', physical_capacity)

# new pool methods
    def get_used_physical_capacity(self):
        # TODO: backward compatibility. should be removed
        return self.get_allocated_physical_capacity()
    def get_allocated_physical_capacity(self):
        return self._get_underlying_object()['allocated_physical_space']
    def get_used_virtual_capacity(self):
        # TODO: backward compatibility. should be removed
        return self.get_allocated_virtual_capacity()
    def get_allocated_virtual_capacity(self):
        return self.get_virtual_capacity() - self.get_free_virtual_capacity()

    def get_free_virtual_capacity(self):
        return self._get_size_field('free_virtual_space')
    def get_free_physical_capacity(self):
        return self._get_size_field('free_physical_space')
    def get_reserved_capacity(self):
        return self._get_size_field('reserved_capacity')
    def get_max_extend(self):
        return self._get_size_field('max_extend')
    def set_max_extend(self, value):
        return self._set_size_field('max_extend', value)
    def get_state(self):
        return self._api.get('state')

    def lock(self):
        self._api.post('lock')
    def unlock(self):
        self._api.post('unlock')
    def is_locked(self):
        return self.get_state() == self.states.locked
    def is_normal(self):
        return self.get_state() == self.states.normal
    def is_limited(self):
        return self.get_state() == self.states.limited
    def get_physical_capacity_critical(self):
        return self._api.get('physical_capacity_critical')
    def get_physical_capacity_warning(self):
        return self._api.get('physical_capacity_warning')
    def set_physical_capacity_critical(self, value):
        return self._api.put('physical_capacity_critical', value)
    def set_physical_capacity_warning(self, value):
        return self._api.put('physical_capacity_warning', value)
    def get_ssd_enabled(self):
        return self._api.get('ssd_enabled')
    def set_ssd_enabled(self, value):
        return self._api.put('ssd_enabled', value)

# old pool methods (some should stay)
    def get_disk_usage_thick(self):
        raise NotImplementedError("deprecated")
        return self._get_size_field('disk_usage_thick')
    def get_disk_usage_thin(self):
        raise NotImplementedError("deprecated")
        return self._get_size_field('disk_usage_thin')
    def get_disk_usage_snap(self):
        raise NotImplementedError("deprecated")
        return self._get_size_field('disk_usage_snap')
    def get_disk_usage_dirty(self):
        raise NotImplementedError("deprecated")
        return self._get_size_field('disk_usage_dirty')
    def get_disk_unused(self):
        return self.get_free_physical_capacity()
    def get_unallocated_capacity(self):
        return self.get_free_physical_capacity()
    def get_thick_size(self):
        raise NotImplementedError("deprecated")
        return self._get_size_field('thick_size')
    def get_volume_count(self):
        return self._api.get('volumes_count')
    def get_entities_count(self):
        return self._api.get('entities_count')
    def get_free_thick_size(self):
        raise NotImplementedError("deprecated")
        return self.get_physical_capacity() - self.get_thick_size()

    def _get_size_field(self, field_name):
        return convert_size_from_system(self._api.get(field_name))
    def _set_size_field(self, field_name, value):
        self._api.put(field_name, convert_size_for_system(value))
    # # operations
    def purge(self):
        mapped_volumes = True
        while mapped_volumes:
            mapped_volumes = self._system.objects.Volume.find(pool_id=self.get_id(), mapped=True, page_size=1000)
            for volume in mapped_volumes:
                volume.unmap()
        master_volumes = True
        while master_volumes:
            master_volumes = self.get_master_volumes(page_size=1000)
            for volume in master_volumes:
                volume.delete()
        super(Pool, self).purge()
    def _safe_get_pool_id(self, volume):
        try:
            return volume.get_pool()
        except APICommandFailed as e:
            if e.ctx.status != httplib.NOT_FOUND:
                raise
            return None
    def _unmap_volumes(self, volumes, mapping_cls, mapping_id, mapping_objects):
        for mapping_object in mapping_objects:
            for lun in mapping_object['luns']:
                volume_id = lun['volume_id']
                volume = Volume(self._system, volume_id)
                pool_id = volumes.get(volume_id, self._safe_get_pool_id(volume))
                if pool_id == self:
                    mapping_cls(self._system, lun[mapping_id]).unmap_volume(volume=volume)
                volumes[volume_id] = pool_id
    def delete(self):
        ctx = Munch(result=None)
        with self.get_system().callbacks.pool_deletion_context.get_context(pool=self, result=ctx):
            super(Pool, self).delete()
    def get_master_volumes(self, **kwargs):
        return self._system.objects.Volume.find(pool_id=self._id, parent_id=0, **kwargs)
    def get_all_volumes(self, **kwargs):
        return self._system.objects.Volume.find(pool_id=self._id, **kwargs)
    def _get_underlying_object(self):
        underlying_object = super(Pool, self)._get_underlying_object()
        if underlying_object.has_key('used_physical_space'):
            underlying_object['allocated_physical_space'] = underlying_object.pop('used_physical_space')
        for capacity_attribute in self.CAPACITY_ATTRIBUTES:
            underlying_object[capacity_attribute] = convert_size_from_system(underlying_object[capacity_attribute])
        return underlying_object
    def add_owner(self, user):
        self._api.post("owners/{}".format(user.get_id()))
    def remove_owner(self, user):
        self._api.delete("owners/{}".format(user.get_id()))
    def list_owners(self):
        return [self._system.types.User(self._system, uid) for uid in self._api.get("owners")]
