import httplib
import logging
import socket
from contextlib import contextmanager
from numbers import Number
from types import NoneType

import waiting
from munch import Munch
from urlobject import URLObject

from infi.pyutils.lazy import cached_method, cached_property
from infinibox_sysdefs import latest as sysdefs

from ..config import config as infinipy_config
from .address_info import (NotSimulator, RealSystemAddressInfo,
                           SimulatorAddressInfo)
from .api import RestAPI
from .callbacks import SystemCallbacks
from .cluster import Cluster
from .configuration import Configuration as SystemConfiguration
from .drive import Drive
from .email_rule import EmailRule
from .enclosure import Enclosure
from .ups import UPS
from .events import Events
from .exceptions import APICommandException, UnkownComponentState
from .host import Host
from .node import Node
from .notification_rule import NotificationRule
from .notification_target import NotificationTarget
from .physical_object import StatefulObject
from .pool import Pool
from .system_capacity import SystemCapacity
from .system_object import CollectionSystemObject
from .system_stats import SystemStats
from .tagging import InMemoryTaggingService
from .users import User, Users
from .version import Version
from .volume import Volume
from .service_cluster import ServiceCluster

DRIVE_STATES = sysdefs.enums.enclosures.drives.states
SYSTEM_STATES = sysdefs.enums.system.states

_logger = logging.getLogger(__name__)

_PYLINT = False


class SystemObjects(object):
    if _PYLINT:
        Cluster = Host = Volume = Pool = ConsGroup = EmailRule = None


class System(StatefulObject):
    types = Munch((objtype.__name__, objtype) for objtype in [
        Cluster,
        Host,
        Pool,
        Version,
        Volume,
        EmailRule,
        User,
        NotificationTarget,
        NotificationRule,
        ])
    states = SYSTEM_STATES

    def __init__(self, api_address_or_simulator, api_port=None, username=None, password=None, use_ssl=False, ssl_cert=None):
        super(System, self).__init__()

        if not isinstance(api_port, (NoneType, Number)):
            raise ValueError(
                "Port number must be a number ({0!r} given)".format(api_port))

        self.callbacks = SystemCallbacks(self)
        self.address_info = self._get_system_address_info(
            api_address_or_simulator, api_port, use_ssl)

        self.simulator = self.address_info.simulator

        self._api_root = RestAPI(self, self.address_info.api_url, username=username,
                                 password=password, resolve_hostname=self.simulator is not None, ssl_cert=None)

        self.api = self._api_root.get_sub_api("/api/rest")
        self.internal_api = self._api_root.get_sub_api("/api/internal")
        self.objects = SystemObjects()
        for (name, objtype) in self.types.iteritems():
            if issubclass(objtype, CollectionSystemObject):
                setattr(self.objects, name, objtype.create_binder(self))
        self._tagging = InMemoryTaggingService()
        self.stats = SystemStats(self)
        self.capacity = SystemCapacity(self)
        self._attached_attributes = set()
        self.events = Events(self)
        self.users = self.objects.User

    def login(self):
        username, password = self.api.get_username_password()
        result = self.api.post("users/login", {"username": username, "password": password})
        # login was successful - make sure system name is cached
        unused = self.get_name()
        user_object = self._find_highest_ranking_user_object(result['user_objects'])
        user_id = user_object['id']
        return User(self, user_id)

    def _find_highest_ranking_user_object(self, user_objects):
        # A hack to deal with the way LDAP is implemented. There may be several user objects returned from login
        # requests, and we want to refer to the highest ranking among them
        ranking = ['infinidat', 'technician', 'admin', 'pooladmin', 'readonly']

        def key_func(user_object):
            return min(ranking.index(role.lower()) for role in user_object['roles'])

        return sorted(user_objects, key=key_func)[0]

    @contextmanager
    def get_changed_floating_ip_context(self, ip_address):
        old_ip_address = self.address_info.ip_address

        def _set(address):
            self.address_info.ip_address = address
            self.address_info.api_url = self.address_info.api_url.with_hostname(address)
            self._api_root.api_url = self.address_info.api_url

        _set(ip_address)
        try:
            yield
        finally:
            _set(old_ip_address)


    _name = None

    def get_name(self, query_system=True):
        returned = self._name
        if returned is None:
            if query_system:
                returned = self._name = self.api.get("system/name")
            else:
                returned = self.address_info.ip_address
        return returned

    name = property(get_name)

    _configuration = None

    @property
    def configuration(self):
        if self._configuration is None:
            self._configuration = SystemConfiguration(self)
        return self._configuration

    def _get_system_address_info(self, api_address_or_simulator, api_port, use_ssl):
        try:
            return SimulatorAddressInfo(api_address_or_simulator, api_port, use_ssl)
        except NotSimulator:
            return RealSystemAddressInfo(api_address_or_simulator, api_port, use_ssl)

    @cached_property
    def _nodes(self):
        node_infos = self.api.get("components/racks/1/nodes")
        return [Node(self, node_info) for node_info in node_infos]

    @cached_method
    def is_virtual(self):
        return self.get_serial() > 30000 and not self.is_simulator()

    @cached_property
    def _enclosures(self):
        return [Enclosure(self, enclosure_info) for enclosure_info in self.api.get("components/racks/1/enclosures")]

    @cached_property
    def _service_clusters(self):
        return [ServiceCluster(self, service_cluster_info) for service_cluster_info in self.api.get("services")]

    @cached_property
    def _upses(self):
        return [UPS(self, ups_info) for ups_info in self.api.get("components/racks/1/ups")]

    def get_tagging_service(self):
        return self._tagging
    tagging = property(get_tagging_service)

    def set_tagging_service(self, service):
        self._tagging = service

    def __repr__(self):
        return str(self)

    def __str__(self):
        return self.get_name(query_system=False)

    def get_fiber_target_addresses(self):
        """Returns addresses of FC ports on the system which serve as targets"""
        return [fiber_port.get_wwpn() for node in self.get_nodes() for fiber_port in node.get_fiber_ports()]

    def get_hostname(self):
        return self.address_info.hostname

    def get_ssd_enabled(self):
        return self.configuration.get_parameter("mgmt/pool.ssd_enabled_default")

    def set_ssd_enabled(self, value):
        return self.configuration.set_parameter("mgmt/pool.ssd_enabled_default", value)

    def get_state(self):
        return self.api.get("system/operational_state")['state']

    def get_deployment_id(self):
        return self.configuration.get_deployment_id()

    def get_clusters(self, *args, **kwargs):
        return self.objects.Cluster.find(*args, **kwargs)

    def get_hosts(self, *args, **kwargs):
        return self.objects.Host.find(*args, **kwargs)

    def get_pools(self, *args, **kwargs):
        return self.objects.Pool.find(*args, **kwargs)

    def get_volumes(self, *args, **kwargs):
        return self.objects.Volume.find(*args, **kwargs)

    def get_volume_groups(self):
        return self.objects.ConsGroup.find()

    def get_mail_rules(self):
        return self.objects.EmailRule.find()

    # Status information
    def get_notification_targets(self):
        return self.objects.NotificationTarget.find()

    def get_enclosures(self):
        return self._enclosures

    def get_enclosure(self, enclosure_index):
        for enclosure in self.get_enclosures():
            if enclosure.get_id() == enclosure_index:
                return enclosure

    def get_service_clusters(self):
        return self._service_clusters

    def get_service_cluster(self, service_cluster_name):
        for service_cluster in self.get_service_clusters():
            if service_cluster.get_name() == service_cluster_name:
                return service_cluster
        return None

    def get_nodes(self):
        """
        :rtype: a list of :class:`infinipy.system.node.Node` objects
        """
        return self._nodes

    def get_running_nodes(self):
        return [node for node in self.get_nodes() if node.is_running()]

    def get_node_by_wwpn(self, wwpn):
        for node in self.get_nodes():
            for fiber_port in node.get_fiber_ports():
                if fiber_port.get_wwpn() == wwpn:
                    return node
        raise LookupError("Cannot find server with wwpn {0}".format(wwpn))

    def get_upses(self):
        return self._upses

    def get_host_id_by_initiator_address(self, address):
        try:
            return self.api.get("hosts/host_id_by_initiator_address/{0}".format(address))
        except APICommandException as e:
            if e.ctx.status != httplib.NOT_FOUND:
                raise
            return None

    def has_registered_initiator_address(self, address):
        return self.get_host_id_by_initiator_address(address) is not None

    def get_administered_pools(self):
        pool_jsons = self.api.get('pools/administered_pools')
        return [Pool(self, pool_json['id']) for pool_json in pool_jsons]

    def get_node_by_id(self, index):
        return self._nodes[index - 1]

    def get_drives(self):
        return [drive for enclosure in self.get_enclosures() for drive in enclosure.get_drives()]

    def get_running_drives(self):
        return self.get_drives_by_states([DRIVE_STATES.active])

    def get_drives_by_states(self, states, ignore_unknown_enclosures=False):
        """ A fast and hacky way to get drives by list of states, using only one API query. """
        component_state = self.api.get("components/racks/1/enclosures")
        if not ignore_unknown_enclosures and any(enclosure_state['state'] == 'UNKOWN' for enclosure_state in component_state):
            raise UnkownComponentState()
        return [drive
                for drive in self.get_drives()
                if (component_state[drive.get_enclosure_index() - 1]['state'] != 'UNKOWN' and
                    component_state[drive.get_enclosure_index() - 1]['drives'][drive.get_drive_index() - 1]['state'] in states)]

    def get_drive(self, enclosure_index, drive_index):
        enclosure = self.get_enclosure(enclosure_index)
        if enclosure:
            return self.get_enclosure(enclosure_index).get_drive_by_index(drive_index)

    def get_serial(self):
        return self.api.get("system/serial_number")

    def get_uptime(self):
        return self.api.get("system/uptime")

    @cached_method
    def get_version(self):
        version = self.api.get("system/version")
        return version

    @cached_method
    def get_revision(self):
        system_info = self.api.get('system')
        returned = system_info.get('release')
        if returned is not None:
            return returned['system']['revision']
        return system_info.get('revision')

    def get_api_port(self):
        return infinipy_config.root.defaults.system_api_port

    def is_simulator(self):
        return self.simulator is not None

    def get_rack_id(self):
        return 1

    def purge(self):
        # Unmap hosts and clusters
        self._unmap_volumes(Cluster, 'host_cluster_id',
                            self.api.get('clusters?page_size=1000'))
        self._unmap_volumes(
            Host, 'host_id', self.api.get('hosts?page_size=1000'))
        # Delete master volumes
        for pool in self.get_pools():
            master_volumes = pool.get_master_volumes()
            while master_volumes:
                for volume in master_volumes:
                    volume.delete()
                master_volumes = pool.get_master_volumes()
        # Delete other objects
        for collection in (self.get_pools(), self.get_hosts(), self.get_clusters()):
            for obj in collection:
                obj.purge()

    def _unmap_volumes(self, mapping_cls, mapping_id, mapping_objects):
        for mapping_object in mapping_objects:
            for lun in mapping_object['luns']:
                if (lun['clustered'] and mapping_cls == Cluster) or not lun['clustered']:
                    volume_id = lun['volume_id']
                    volume = Volume(self, volume_id)
                    mapping_cls(self, lun[mapping_id]
                                ).unmap_volume(volume=volume)

    def is_active(self):
        try:
            return super(System, self)._is_state(self.states.active)
        except APICommandException:
            return False
        except socket.error:
            return False

    def is_stand_by(self):
        try:
            return super(System, self)._is_state(self.states.standby)
        except APICommandException:
            return False
        except socket.error:
            return False

    def is_down(self):
        try:
            super(System, self)._is_state(self.states.active)
            return False
        except APICommandException:
            return True
        except socket.error:
            return True

    def shutdown(self, emergency=False, wait=True, wake_in_stand_by=False, power_off=False):
        # make sure nodes are initialized - we won't have a chance to query
        # them once the system goes down
        unused = self._nodes
        if wake_in_stand_by:
            self.configuration.set_parameter(
                "mgmt/state.pause_startup_in", "STANDBY")
        shutdown_type = "emergency" if emergency else "normal"
        _logger.info("Shutting down %s (type - %s)", self, shutdown_type)
        if emergency:
            if power_off:
                self.configuration.set_parameter("mgmt/state.poweroff", True)
            self.api.post('system/operational_state/emergency_shutdown')
        else:
            if power_off:
                self.api.post('system/operational_state/poweroff')
            else:
                self.api.post('system/operational_state/shutdown')
        if wait:
            _logger.info("Waiting for %s API to shut down...", self)
            self._wait_for_shutdown()
            _logger.info("%s API is down.", self)

    def _wait_for_shutdown(self):
        waiting.wait(self.is_down,
                     timeout_seconds=infinipy_config.root.timeouts.system_shutdown)

    def search(self, name_pattern, obj_type=None):
        query = URLObject("search").add_query_param("query", name_pattern)
        if obj_type is not None:
            query = query.add_query_param("type", obj_type)

        cls_by_type = lambda name: getattr(self.types, name.capitalize())
        return [cls_by_type(obj['type'])(self, obj["id"]) for obj in self.api.get(query)]
