import random
import httplib
from urlobject import URLObject
from infi.pyutils import renumerate
from infi.pyutils.predicates import DictionaryItems
from .exceptions import (
    AmbiguousCriteria,
    APICommandFailed,
    NoObjectFound,
    )
from .taggable_object import TaggableObject
from .system_object_type_binder import SystemObjectTypeBinder
from capacity import Capacity, byte
from sentinels import NOTHING as ALL_METADATA


class SystemObject(object):
    def __init__(self, system):
        super(SystemObject, self).__init__()
        self._system = system

    @classmethod
    def create_binder(cls, system):
        return SystemObjectTypeBinder(system, cls)

    @classmethod
    def get_binder_bound_method_names(cls):
        return []

    def get_system(self):
        return self._system
    def get_resource_path(self):
        raise NotImplementedError()
    _cached_api = None
    @property
    def _api(self):
        returned = self._cached_api
        if returned is None:
            returned = self._cached_api = self._system.api.get_sub_api(self.get_resource_path())
        return returned
    def is_in_system(self):
        try:
            self._api.get()
        except APICommandFailed as e:
            if e.ctx.status != httplib.NOT_FOUND:
                raise
            return False
        return True
    def __eq__(self, other):
        if not isinstance(other, SystemObject):
            return False
        return self._get_comparison_key() == other._get_comparison_key()
    def __ne__(self, other):
        return not (self == other)
    def __hash__(self):
        return hash(self._get_comparison_key())
    def _get_comparison_key(self):
        raise NotImplementedError()

class CollectionSystemObject(SystemObject, TaggableObject):
    # contains the list of field names that can be searched via the query parameters
    # TODO: this should be attached to attributes later on
    SEARCHABLE_FIELD_NAMES = set(["name", "name_contains"])
    def __init__(self, system, id):
        super(CollectionSystemObject, self).__init__(system)
        self._id = id
    def get_id(self):
        return self._id
    def _get_tag_id(self):
        return self._id
    @classmethod
    def get_collection_path(cls):
        return cls.get_object_type_name() + 's'
    @classmethod
    def get_object_type_name(cls):
        return cls.__name__.lower()
    @classmethod
    def _create_from_post(cls, system, data):
        result = system.api.post(cls.get_collection_path(), data)
        return cls(system, result['id'])
    def get_resource_path(self):
        return '/'.join([self.get_collection_path(), str(self._id)])
    def _get_comparison_key(self):
        return (type(self), self._system, self._id)
    def _get_underlying_object(self, sub_url=None):
        return self._api.get(sub_url)
    @classmethod
    def _get_underlying_response(cls, system, page=None, page_size=None, **fields):
        searchable, unsearchable = cls._separate_searchable_fields(fields)
        uri = URLObject(cls.get_collection_path())
        if page_size is None:
            page_size = 1000
        if page is None:
            page = 1
        for param, value in searchable.iteritems():
            uri = uri.add_query_param(param, "eq:{0}".format(cls._externalize_value(value)))

        if page is not None:
            uri = uri.add_query_param("page", str(page))
        if page_size is not None:
            uri = uri.add_query_param("page_size", str(page_size))
        returned = system.api.get(str(uri), raw_json_response=True)
        return unsearchable, returned
    @classmethod
    def _get_all_underlying_objects(cls, system, page=None, page_size=None, **fields):
        unsearchable, returned = cls._get_underlying_response(system, page, page_size, **fields)
        returned = returned["result"]
        # <workaround>
        # TODO: this is to deal with INFINIBOX-5227
        if "name" in fields:
            unsearchable.update(name=fields["name"])
        # </workaround>
        if unsearchable:
            predicate = DictionaryItems(**unsearchable)
            returned = [obj for obj in returned if predicate(obj)]
        return returned
    @classmethod
    def _separate_searchable_fields(cls, fields):
        searchable = {}
        unsearchable = {}
        for field_name, field_value in fields.iteritems():
            (searchable if field_name in cls.SEARCHABLE_FIELD_NAMES else unsearchable)[field_name] = field_value
        return searchable, unsearchable
    @classmethod
    def _externalize_value(cls, value):
        if isinstance(value, Capacity):
            return long(value // byte)
        return value
    @classmethod
    def find(cls, system, *args, **kwargs):
        if not args and len(kwargs) == 1 and "id" in kwargs:
            return cls._find_by_id(system, kwargs["id"])
        return [cls(system, obj["id"]) for obj in cls._get_all_underlying_objects(system, *args, **kwargs)]

    @classmethod
    def _find_by_id(cls, system, id):
        returned = cls(system, id)
        if not returned.is_in_system():
            raise NoObjectFound("{0}[id={1}] not found".format(cls.__name__, id))
        return [returned]

    @classmethod
    def count(cls, system,):
        return cls._get_underlying_response(system, page_size=1)[1]['metadata']['number_of_objects']
    @classmethod
    def search(cls, system, query):
        type_name = cls.get_object_type_name()
        query = URLObject("search").add_query_param("type", type_name).add_query_param("query", query)
        return [
            cls(system, obj["id"])
            for obj in system.api.get(query)
            ]
    @classmethod
    def get(cls, system, **fields):
        objects = cls.find(system, **fields)
        if not objects:
            raise NoObjectFound("No object matched criteria: {0}".format(fields))
        if len(objects) > 1:
            raise AmbiguousCriteria("Multiple {0} objects matched criteria, which should've been unique: {1}"
                                    .format(cls.__name__.lower(), fields))
        return objects[0]
    @classmethod
    def safe_get(cls, system, **fields):
        try:
            objects = cls.find(system, **fields)
            return objects[0] if len(objects) == 1 else None
        except NoObjectFound:
            return None
    @classmethod
    def choose(cls, system, **fields):
        objects = cls.find(system, **fields)
        return random.choice(objects) if objects else None
    @classmethod
    def recreate(cls, system, name, **fields):
        for obj in cls.find(system, name=name):
            obj.purge()
        return cls.create(system, name=name, **fields)
    def get_metadata(self, key=ALL_METADATA):
        if key is ALL_METADATA:
            return self.get_all_metadata()
        return self._system.api.get('{0}/{1}'.format(self._get_metadata_uri(), key))
    def get_all_metadata(self):
        return self._system.api.get(self._get_metadata_uri())
    def set_metadata(self, *args, **kwargs):
        if args:
            kwargs.update({args[0]:args[1]})
        return self._system.api.put(self._get_metadata_uri(), kwargs)
    def unset_metadata(self, key):
        return self._system.api.delete("{0}/{1}".format(self._get_metadata_uri(), key))
    def clear_metadata(self):
        return self._system.api.delete(self._get_metadata_uri())
    def _get_metadata_uri(self):
        return 'metadata/{0}'.format(self.get_id())
    def get_name(self):
        return self._api.get('name')
    def set_name(self, new_name):
        self._api.put('name', new_name)
    def delete(self):
        self._api.delete()
        self.remove_tags()
    def purge(self):
        self.delete()
    def __repr__(self):
        return "<{0} {1!r}:{2}>".format(self.__class__.__name__, self.get_system(), self.get_id())


class NewAPICollectionSystemObject(CollectionSystemObject):
    # objects which support only new API (no direct get/put)
    def update_fields(self, **fields):
        for k, v in fields.items():
            self._api.put('', data={k:v})
    def update_field(self, field, value):
        return self.update_fields(**{field:value})
    def get_fields(self, *fields):
        data = self._get_underlying_object()
        if fields:
            data = dict((field, data[field]) for field in fields)
        return data
    def get_field(self, field):
        return self.get_fields(field)[field]
    def get_name(self):
        return self.get_field('name')
    def set_name(self, new_name):
        return self.update_field('name', new_name)
