from .exceptions import NoSuchTag
from sentinels import Sentinel

NO_VALUE = Sentinel("NO_VALUE")
NO_DEFAULT = Sentinel("NO_DEFAULT")

class TaggingService(object):
    def _get_raw_tags(self, id):
        raise NotImplementedError
    def _get_raw_tag(self, id, key):
        tags = self._get_raw_tags(id)
        return tags.get(key, NO_VALUE) if tags else NO_VALUE
    def get_tags(self, id):
        tags = self._get_raw_tags(id)
        return dict(tags) if tags is not None else {}
    def get_tag(self, id, key, default=NO_DEFAULT):
        value = self._get_raw_tag(id, key)
        if value is not NO_VALUE:
            return value
        if default is not NO_DEFAULT:
            return default
        raise NoSuchTag(key)
    def set_tag(self, id, key, value):
        self.update_tags(id, **{key:value})
    def remove_tag(self, id, key):
        raise NotImplementedError
    def remove_tags(self, id):
        raise NotImplementedError
    def update_tags(self, id, **tags):
        raise NotImplementedError

class InMemoryTaggingService(TaggingService):
    def __init__(self):
        super(InMemoryTaggingService, self).__init__()
        self._tags = {}
    def _get_raw_tags(self, id):
        return self._tags.get(id)
    def remove_tag(self, id, key):
        tags = self._get_raw_tags(id)
        if tags:
            value = tags.pop(key, NO_VALUE)
            if value is not NO_VALUE:
                return
        raise NoSuchTag(key)
    def remove_tags(self, id):
        self._tags.pop(id, None)
    def update_tags(self, id, **tags):
        curr_tags = self._get_raw_tags(id)
        if curr_tags is None:
            curr_tags = {}
            self._tags[id] = curr_tags
        curr_tags.update(**tags)

class MongoDBTaggingService(TaggingService):
    ID_FIELD, KEY_FIELD, VALUE_FIELD = "object_id", "key", "value"
    SYSTEM_NAME_FIELD, SYSTEM_UUID_FIELD = "system_name", "system_uuid"
    def __init__(self, system, collection):
        super(MongoDBTaggingService, self).__init__()
        self.system = system
        self.collection = collection
        self._system_identification_criteria = {
            self.SYSTEM_NAME_FIELD : self.system.get_name(),
            self.SYSTEM_UUID_FIELD : self.system.get_deployment_id(),
            }
    def _get_criteria(self, id=None, key=None):
        criteria = dict(self._system_identification_criteria)
        if id:
            criteria[self.ID_FIELD] = id
        if key:
            criteria[self.KEY_FIELD] = key
        return criteria
    def _get_raw_tags(self, id):
        tag_records = self.collection.find(self._get_criteria(id))
        return dict((tag[self.KEY_FIELD],  tag[self.VALUE_FIELD]) for tag in tag_records)
    def _get_raw_tag(self, id, key):
        tag_record = self.collection.find_one(self._get_criteria(id, key))
        if tag_record is None:
            return NO_VALUE
        return tag_record["value"]
    def get_tags(self, id):
        return self._get_raw_tags(id)
    def remove_tag(self, id, key):
        self.collection.remove(self._get_criteria(id, key))
        lasterror = self.collection.database.command("getlasterror")
        if lasterror.get("n", 0) <= 0:
            raise NoSuchTag(key)
    def remove_tags(self, id):
        criteria = self._get_criteria(id)
        criteria["$atomic"] = True
        self.collection.remove(criteria)
    def update_tags(self, id, **tags):
        for key, value in tags.iteritems():
            self.collection.update(self._get_criteria(id, key),
                                   {"$set" : {self.VALUE_FIELD : value}},
                                   upsert=True, safe=True)
    def remove_tags_unaffiliated_to_deployment(self):
        self.collection.remove({
            self.SYSTEM_NAME_FIELD : self.system.get_name(),
            self.SYSTEM_UUID_FIELD : { "$ne" : self.system.get_deployment_id() },
            })
