import itertools
import logging
from flux import current_timeline as time
from contextlib import contextmanager
from contextlib2 import ExitStack
from munch import Munch
from infi.pyutils.lazy import cached_method
from infinibox_sysdefs import latest as sysdefs
from .exceptions import InvalidOperationException
from .system_object import CollectionSystemObject
from .utils import convert_size_for_system
from .utils import convert_size_from_system
from .defaults import (
    has_defaults,
    CapacityFromConfiguration as FROM_CONFIG,
    GENERATE_NAME,
    GENERATE_NAME_WITH_PREFIX,
    )

_logger = logging.getLogger(__name__)

PROVISIONING = sysdefs.enums.volumes.provisioning_types
VOLUME_TYPE = sysdefs.enums.volumes.types

class Volume(CollectionSystemObject):
    SEARCHABLE_FIELD_NAMES = CollectionSystemObject.SEARCHABLE_FIELD_NAMES | set(["parent_id", "pool_id", "mapped", "size"])
    CAPACITY_ATTRIBUTES = ['size', 'allocated', 'tree_allocated', 'used']

    @classmethod
    def get_binder_bound_method_names(cls):
        return ["create_group_snapshot", 'run_delete_simulation']

    @classmethod
    @has_defaults
    def create(cls, system, pool, name=GENERATE_NAME, size=FROM_CONFIG('defaults.volume.size'), provisioning=PROVISIONING.thick):
        ctx = Munch(result=None)
        with system.callbacks.volume_creation_context.get_context(pool=pool, name=name, size=size, provisioning=provisioning, result=ctx):
            ctx.result = returned = cls._create_from_post(system,
                                                          dict(
                                                               name=name,
                                                               size=convert_size_for_system(size),
                                                               pool_id=pool.get_id(),
                                                               provtype=str(provisioning)
                                                               ))
            return returned


    @classmethod
    def create_group_snapshot(cls, system, volumes, prefix=None, suffix=""):
        """
        Creates a snapshot group using the group consistent snapshot API.

        :param volumes: either a list of volumes or a dict mapping volumes to their snapshot names. Snapshot names can be None meaning default
        """

        if prefix is None:
            prefix = str(int(time.time() * 100))[-5:]


        entities = []
        if isinstance(volumes, (dict)):
            volume_iterator = volumes.iteritems()
        else:
            volume_iterator = itertools.izip(volumes, itertools.repeat(None))

        volumes_and_names = list(volume_iterator)
        contexts = []

        result = []

        with ExitStack() as stack:
            for volume, snapshot_name in volumes_and_names:
                contexts.append(stack.enter_context(volume._get_snapshot_context()))
            snapshot_infos = system.api.post("volumes/group_snapshot", {
                "entities": [{"id": v.get_id(), "snap_name": name} for v, name in volumes_and_names],
                "snap_prefix": prefix, "snap_suffix": suffix})
            assert len(snapshot_infos) == len(volumes_and_names), "Mismatch between volume list length and snapshot list length"
            snapshots_by_parent_id = dict((snapshot_info["parent_id"], snapshot_info) for snapshot_info in snapshot_infos)
            for ctx, (volume, _) in zip(contexts, volumes_and_names):
                ctx.result = Volume(system, snapshots_by_parent_id[volume.get_id()]["id"])
                result.append(ctx.result)

        return result


    @classmethod
    def run_delete_simulation(cls, system, volumes):
        data = system.api.post("volumes/delete_simulation", data=dict(entities=[v.get_id() for v in volumes]))
        data['space_reclaimable'] = convert_size_from_system(data['space_reclaimable'])
        return data

    def get_unique_key(self):
        return self._get_comparison_key()

    # # attributes
    def get_pool(self):
        return self._system.objects.Pool(self._api.get('pool_id'))
    def set_pool(self, new_pool, with_capacity=False):
        with self.get_system().callbacks.volume_move_context.get_context(volume=self, pool=new_pool):
            self._api.post('move', {"pool_id": new_pool.get_id(), "with_capacity": with_capacity})
    def get_hosts(self):
        return self._system.objects.Host.find(volume_id=self._id)
    def get_provisioning(self):
        return PROVISIONING.get(self._api.get('provtype'))
    def set_provisioning(self, provisioning):
        return PROVISIONING.get(self._api.put('provtype', str(provisioning)))
    def get_serial(self):
        return self._api.get("serial")
    def get_size(self):
        return convert_size_from_system(self._api.get('size'))
    def set_size(self, size):
        ctx = Munch(result=None)
        with self.get_system().callbacks.volume_resize_context.get_context(volume=self, size=size, result=ctx):
            ctx.result = returned = self._api.put('size', convert_size_for_system(size))
            return returned
    def get_children(self):
        return self.get_system().objects.Volume.find(parent_id=self._id)
    def get_snapshots(self):
        if self.is_snapshot():
            return []
        return self.get_children()
    def get_clones(self):
        if not self.is_snapshot():
            return []
        return self.get_children()
    def get_write_protected(self):
        return self._api.get('write_protected')
    def set_write_protected(self, value):
        return self._api.put('write_protected', value)
    @cached_method
    def get_type(self):
        return VOLUME_TYPE.get(self._api.get('type'))
    def get_parent(self):
        parent_id = self.get_parent_id()
        if parent_id is None:
            return None
        return self._system.objects.Volume(parent_id)
    def get_parent_id(self):
        return self._api.get('parent_id') or None
    def get_luns(self):
        return self._api.get("luns")
    def is_master_volume(self):
        return VOLUME_TYPE.master == self.get_type()

    def is_mapped(self):
        return self._api.get("mapped")

    def is_snapshot(self):
        return VOLUME_TYPE.snap == self.get_type()
    def is_clone(self):
        return VOLUME_TYPE.clone == self.get_type()
    def get_depth(self):
        return self._api.get('depth')
    def get_allocated_size(self):
        return self._get_underlying_object()['allocated']
    def get_tree_allocated_size(self):
        return self._get_underlying_object()['tree_allocated']
    def get_used_size(self):
        return self._get_underlying_object()['used']
    def get_num_blocks(self):
        return self._api.get('num_blocks')
    def is_thick(self):
        return PROVISIONING.thick == self.get_provisioning()
    def is_thin(self):
        return PROVISIONING.thin == self.get_provisioning()
    def has_children(self):
        return self._api.get('has_children')
    def get_ssd_enabled(self):
        return self._api.get('ssd_enabled')
    def set_ssd_enabled(self, value):
        return self._api.put('ssd_enabled', value)

    # # operations
    @has_defaults
    def create_clone(self, name=GENERATE_NAME_WITH_PREFIX("clone")):
        return self._create_clone_or_snapshot(name, snapshot=False)
    @has_defaults
    def create_snapshot(self, name=GENERATE_NAME_WITH_PREFIX("snap")):
        return self._create_clone_or_snapshot(name, snapshot=True)
    def _create_clone_or_snapshot(self, name, snapshot):
        if snapshot and self.is_snapshot():
            raise InvalidOperationException("Cannot create a snapshot for a snapshot")
        if not snapshot and not self.is_snapshot():
            raise InvalidOperationException("Cannot create a clone for a volume or a clone")

        with self._get_snapshot_context() as ctx:
            ctx.result = returned = self._create_from_post(self.get_system(), dict(name=name, parent_id=self.get_id()))
            _logger.debug("Created %s %s --> %s", "snapshot" if snapshot else "clone", self, returned)
        return returned

    @contextmanager
    def _get_snapshot_context(self):
        ctx = Munch(result=None)
        with self.get_system().callbacks.snapshot_creation_context.get_context(volume=self, result=ctx):
            yield ctx

    def purge(self):
        self.unmap()
        for child in self.get_children():
            child.purge()
        super(Volume, self).purge()
    def delete(self):
        ctx = Munch(result=None)
        with self.get_system().callbacks.volume_deletion_context.get_context(volume=self, result=ctx):
            super(Volume, self).delete()
    def unmap(self):
        "Unmaps this volume from all hosts and clusters"
        for lun_info in self.get_luns():
            if lun_info['clustered']:
                mapping_object = self._system.objects.Cluster.get(id=lun_info['host_cluster_id'])
            else:
                mapping_object = self._system.objects.Host.get(id=lun_info['host_id'])
            mapping_object.unmap_volume(self, force=True)
    def restore(self, snapshot):
        _logger.debug("Restoring %s from %s", self, snapshot)
        with self._system.callbacks.snapshot_restore_context.get_context(volume=self, snapshot=snapshot):
            self._api.post("restore", snapshot.get_id())

    def get_family_master(self):
        ''' get master volume of any volume / snapshot / clone. if volume is master it will be returned '''
        volume = self
        parent = self.get_parent()
        while (parent):
            volume = parent
            parent = volume.get_parent()
        return volume

    def get_subtree(self):
        ''' Returns a list of all volumes in sub-tree below self (not including self)'''
        returned = []
        parent_list = [self]
        while parent_list:
            parent_list = [child for member in parent_list for child in member.get_children()]
            returned.extend(parent_list)
        return returned

    def _get_underlying_object(self):
        underlying_object = super(Volume, self)._get_underlying_object()
        if underlying_object.has_key('referenced'):
            underlying_object['used'] = underlying_object.pop('referenced')
        for capacity_attribute in self.CAPACITY_ATTRIBUTES:
            underlying_object[capacity_attribute] = convert_size_from_system(underlying_object[capacity_attribute])
        return underlying_object

    def own_replication_snapshot(self, snapshot_name):
        snapshot_id = self._api.post('own_snapshot', {'name': snapshot_name})['id']
        return Volume(self.get_system(), snapshot_id)
